# gemato: hash support
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license

import hashlib
import io

import gemato.exceptions


HASH_BUFFER_SIZE = 65536


class SizeHash(object):
	"""
	A cheap wrapper to count file size via hashlib-like interface.
	"""

	__slots__ = ['size']

	def __init__(self):
		self.size = 0

	def update(self, data):
		self.size += len(data)

	def hexdigest(self):
		return self.size


def get_hash_by_name(name):
	"""
	Get a hashlib-compatible hash object for hash named @name. Supports
	multiple backends.
	"""
	# special case hashes
	if name == '__size__':
		return SizeHash()

	# general hash support
	if name in hashlib.algorithms_available:
		return hashlib.new(name)

	# fallback support
	if name.startswith('sha3_'):
		try:
			import sha3
		except ImportError:
			pass
		else:
			try:
				return getattr(sha3, name)()
			except AttributeError:
				pass
	elif name.startswith('blake2'):
		try:
			import pyblake2
		except ImportError:
			pass
		else:
			try:
				return getattr(pyblake2, name)()
			except AttributeError:
				pass

	raise gemato.exceptions.UnsupportedHash(name)


def hash_file(f, hash_names):
	"""
	Hash the contents of file object @f using all hashes specified
	as @hash_names. Returns a dict of (hash_name -> hex value) mappings.
	"""
	hashes = {}
	for h in hash_names:
		hashes[h] = get_hash_by_name(h)
	for block in iter(lambda: f.read1(HASH_BUFFER_SIZE), b''):
		for h in hashes.values():
			h.update(block)
	return dict((k, h.hexdigest()) for k, h in hashes.items())


def hash_path(path, hash_names):
	"""
	Hash the contents of file at specified path @path using all hashes
	specified as @hash_names. Returns a dict of (hash_name -> hex value)
	mappings.
	"""
	with io.open(path, 'rb') as f:
		return hash_file(f, hash_names)


def hash_bytes(buf, hash_name):
	"""
	Hash the data in provided buffer @buf using the hash @hash_name.
	Returns the hex value.
	"""
	return hash_file(io.BytesIO(buf), (hash_name,))[hash_name]



#!/usr/bin/env python

import os.path
import sys
import timeit

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

import gemato.hash


def benchmark_one(path, hashes):
	f = lambda: gemato.hash.hash_path(path, hashes)
	t = timeit.repeat(f, number=1)
	print("{} -> {}".format(hashes, t))


def benchmark(path, hash_sets):
	if not hash_sets:
		hash_sets = [
			["sha256"],
			["sha256", "blake2b"],
			["sha256", "blake2b", "sha512"],
		]
	else:
		hash_sets = [x.split() for x in hash_sets]

	for hashes in hash_sets:
		benchmark_one(path, hashes)


if __name__ == '__main__':
	if len(sys.argv) < 2:
		print('Usage: {} <test-file> [<hash-set>...]'.format(sys.argv[0]))
		sys.exit(1)

	benchmark(sys.argv[1], sys.argv[2:])



# gemato: Test utility functions
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license

import io
import logging
import os
import os.path
import shutil
import sys
import tempfile
import unittest


class LoggingTestCase(unittest.TestCase):
    def setUp(self):
        if sys.hexversion < 0x03000000:
            self.log = io.BytesIO()
        else:
            self.log = io.StringIO()
        self.log_handler = logging.getLogger().addHandler(
                logging.StreamHandler(self.log))

    def tearDown(self):
        # TODO: make some use of the log output?
        logging.getLogger().removeHandler(self.log_handler)


class TempDirTestCase(LoggingTestCase):
    DIRS = []
    FILES = {}

    def setUp(self):
        super(TempDirTestCase, self).setUp()
        self.dir = tempfile.mkdtemp()
        for k in self.DIRS:
            os.mkdir(os.path.join(self.dir, k))
        for k, v in self.FILES.items():
            with io.open(os.path.join(self.dir, k), 'w', encoding='utf8') as f:
                f.write(v)

    def tearDown(self):
        shutil.rmtree(self.dir)
        super(TempDirTestCase, self).tearDown()



# gemato: Recursive loader for Manifests
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license

import errno
import multiprocessing
import os.path
import sys

import gemato.compression
import gemato.exceptions
import gemato.manifest
import gemato.profile
import gemato.util
import gemato.verify


class ManifestLoader(object):
    """
    Helper class to load Manifests in subprocesses.
    """

    __slots__ = ['root_directory', 'verify_openpgp', 'openpgp_env']

    def __init__(self, root_directory, verify_openpgp, openpgp_env):
        """
        @root_directory specifies top directory of Manifest tree.

        If @verify_openpgp is True and a Manifest contain an OpenPGP
        signature, the signature will be verified. @openpgp_env
        is the OpenPGP environment to use.
        """
        self.root_directory = root_directory
        self.verify_openpgp = verify_openpgp
        self.openpgp_env = openpgp_env

    def verify_and_load(self, relpath, verify_entry=None):
        """
        Load the Manifest from file @relpath (relative to
        root_directory). If the file is compressed, it is decompressed
        transparently.

        If @verify_entry is not None, the Manifest file is verified
        against the entry. If the verification fails, ManifestMismatch
        exception is raised.

        Returns a tuple of (ManifestFile instance, file stat result).
        """
        m = gemato.manifest.ManifestFile()
        path = os.path.join(self.root_directory, relpath)

        if verify_entry is not None:
            ret, diff = gemato.verify.verify_path(path, verify_entry)
            if not ret:
                raise gemato.exceptions.ManifestMismatch(
                        relpath, verify_entry, diff)

        with gemato.compression.open_potentially_compressed_path(
                path, 'r', encoding='utf8') as f:
            m.load(f, self.verify_openpgp, self.openpgp_env)
            st = os.fstat(f.fileno())

        return m, st

    def __call__(self, args):
        """
        Load the Manifest file by passing @args to verify_and_load()
        method. @args should be an iterable specifying the file relative
        path and verification entry (or None).

        Returns a tuple of (relpath, ManifestFile instance).

        Intended to be used via multiprocessing.Pool.map().
        """
        return (args[0], self.verify_and_load(*args)[0])


class SubprocessVerifier(object):
    """
    Helper class used to verify directories in subprocesses.
    """

    __slots__ = ['top_level_manifest_filename',
                 'manifest_device', 'fail_handler', 'last_mtime']

    def __init__(self, top_level_manifest_filename,
                 manifest_device, fail_handler, last_mtime):
        self.top_level_manifest_filename = top_level_manifest_filename
        self.manifest_device = manifest_device
        self.fail_handler = fail_handler
        self.last_mtime = last_mtime

    def _verify_one_file(self, path, relpath, e):
        ret, diff = gemato.verify.verify_path(path, e,
                expected_dev=self.manifest_device,
                last_mtime=self.last_mtime)

        if not ret:
            err = gemato.exceptions.ManifestMismatch(relpath, e, diff)
            ret = self.fail_handler(err)
            if ret is None:
                ret = True

        return ret

    def __call__(self, vals):
        """
        Verify the specified directory and return the boolean value
        (or raise an exception).
        """

        ret = True
        dirpath, relpath, dirnames, filenames, dirdict = vals

        for d in dirnames:
            # we already stripped ignored directories in walker,
            # so go straight for verification
            de = dirdict.pop(d, None)
            if de is not None:
                dpath = os.path.join(relpath, d)
                ret &= self._verify_one_file(os.path.join(dirpath, d),
                        dpath, de)

        for f in filenames:
            # skip dotfiles
            if f.startswith('.'):
                continue

            fpath = os.path.join(relpath, f)
            # skip top-level Manifest, we obviously can't have
            # an entry for it
            if fpath == self.top_level_manifest_filename:
                continue
            fe = dirdict.pop(f, None)
            ret &= self._verify_one_file(os.path.join(dirpath, f),
                    fpath, fe)


        # check for missing files
        for f, e in dirdict.items():
            fpath = os.path.join(relpath, f)
            ret &= self._verify_one_file(os.path.join(dirpath, f),
                    fpath, e)

        return ret


class ManifestRecursiveLoader(object):
    """
    A class encapsulating a tree covered by multiple Manifests.
    Automatically verifies and loads additional sub-Manifests,
    and provides methods to access the entries in them.
    """

    __slots__ = [
        # configuration properties
        'root_directory',
        'openpgp_env',
        'sign_openpgp',
        'openpgp_keyid',
        'hashes',
        'openpgp_signed',
        'openpgp_signature',
        'sort',
        'compress_watermark',
        'compress_format',
        'profile',
        # internal variables
        'manifest_loader',
        'top_level_manifest_filename',
        'loaded_manifests',
        'updated_manifests',
        'manifest_device',
        'max_jobs',
    ]

    def __init__(self, top_manifest_path,
            verify_openpgp=None, openpgp_env=None,
            sign_openpgp=None, openpgp_keyid=None,
            hashes=None, allow_create=False, sort=None,
            compress_watermark=None, compress_format=None,
            profile=gemato.profile.DefaultProfile(),
            max_jobs=None):
        """
        Instantiate the loader for a Manifest tree starting at top-level
        Manifest @top_manifest_path.

        @verify_openpgp and @openpgp_env are passed down
        to ManifestFile. If the top-level Manifest is OpenPGP-signed
        and the verification succeeds, openpgp_signed property
        is set to True and openpgp_signature will contain the signature
        data. @verify_openpgp is True by default.

        @sign_openpgp is passed down to ManifestFile when writing
        the top-level Manifest. If it is True, the top-level Manifest
        will be signed. If it is False, it will not be signed.
        If it is left as None, then it will be signed if it was
        originally signed. @openpgp_keyid can be used to select the key.

        Sub-Manifests are never signed.

        @hashes can be used to specify a default hash set
        for the Manifest. If it is specified, they will be used for all
        subsequent update*() calls that do not specify another set
        of hashes explicitly.

        If @allow_create is True and @top_manifest_path does not exist,
        a new Manifest tree will be initialized. Otherwise, opening
        a non-existing file will cause an exception.

        If @sort is True, the Manifest entries will be sorted prior
        to saving. By default they are not.

        If @compress_watermark is not None, then the uncompressed
        Manifest files whose size is larger than or equal to the value
        will be compressed using @compress_format. The Manifest files
        whose size is smaller will be uncompressed. To compress all
        Manifest files, pass a size of 0.

        If @compress_watermark is None, the compression is left as-is.
        The default @compress_format is 'gz'.

        @profile can be used to provide the profile for the repository.

        @max_jobs defines the number of subprocesses that can be spawned
        to optimize some operations. If None (the default), the number
        will automatically be determined based on CPU count. Otherwise,
        the specified number will be used.
        """

        self.root_directory = os.path.dirname(top_manifest_path)
        self.openpgp_env = openpgp_env
        self.sign_openpgp = sign_openpgp
        self.openpgp_keyid = openpgp_keyid
        self.hashes = hashes
        self.profile = profile
        self.sort = sort
        self.compress_watermark = compress_watermark
        self.compress_format = compress_format
        self.max_jobs = max_jobs

        self.profile.set_loader_options(self)

        if verify_openpgp is None:
            verify_openpgp = True
        if self.sort is None:
            self.sort = False
        if self.compress_format is None:
            self.compress_format = 'gz'

        self.manifest_loader = ManifestLoader(self.root_directory,
                verify_openpgp, self.openpgp_env)
        self.top_level_manifest_filename = os.path.basename(
                top_manifest_path)
        self.loaded_manifests = {}
        self.updated_manifests = set()

        # TODO: allow catching OpenPGP exceptions somehow?
        m = self.load_manifest(self.top_level_manifest_filename,
                allow_create=allow_create)
        self.openpgp_signed = m.openpgp_signed
        self.openpgp_signature = m.openpgp_signature

    def load_manifest(self, relpath, verify_entry=None,
            allow_create=False):
        """
        Load a single Manifest file whose relative path within Manifest
        tree is @relpath. If @verify_entry is not null, the Manifest
        file is verified against the entry. If the file is compressed,
        it is decompressed transparently.

        If @allow_create is True and the Manifest does not exist,
        a new Manifest will be added. Otherwise, opening a non-existing
        file will cause an exception.
        """

        try:
            m, st = self.manifest_loader.verify_and_load(
                    relpath, verify_entry)
        except IOError as err:
            if err.errno == errno.ENOENT and allow_create:
                m = gemato.manifest.ManifestFile()
                path = os.path.join(self.root_directory, relpath)
                st = os.stat(os.path.dirname(path))
                # trigger saving
                self.updated_manifests.add(relpath)

                # add initial IGNORE entries to top-level Manifest
                if relpath == 'Manifest':
                    for ip in (self.profile
                               .get_ignore_paths_for_new_manifest('')):
                        ie = gemato.manifest.ManifestEntryIGNORE(ip)
                        m.entries.append(ie)
            else:
                raise err

        self.manifest_device = st.st_dev
        self.loaded_manifests[relpath] = m
        return m

    def save_manifest(self, relpath, sort=False):
        """
        Save a single Manifest file whose relative path within Manifest
        tree is @relpath. The Manifest must already be loaded.
        If the name indicates compression, it will be compressed
        transparently. If it was OpenPGP-signed, a new signature
        will be created.

        If @sort is True, the Manifest entries will be sorted prior
        to saving.

        Returns the uncompressed size of the Manifest (number
        of characters written).
        """
        m = self.loaded_manifests[relpath]
        path = os.path.join(self.root_directory, relpath)

        # is it top-level Manifest?
        if relpath == self.top_level_manifest_filename:
            sign = self.sign_openpgp
        else:
            sign = False

        with gemato.compression.open_potentially_compressed_path(
                path, 'w', encoding='utf8') as f:
            m.dump(f, sign_openpgp=sign, sort=sort,
                    openpgp_env=self.openpgp_env,
                    openpgp_keyid=self.openpgp_keyid)
            return f.buffer.tell()

    def _iter_unordered_manifests_for_path(self, path, recursive=False):
        """
        Iterate over loaded Manifests that can apply to path.
        If @recursive is True, returns also Manifests for subdirectories
        of @path. Yields a tuple of (manifest_path, dir_path, manifest).

        The entries will be returned in any order.
        """
        for k, v in self.loaded_manifests.items():
            d = os.path.dirname(k)
            if gemato.util.path_starts_with(path, d):
                yield (k, d, v)
            elif recursive and gemato.util.path_starts_with(d, path):
                yield (k, d, v)

    def _iter_manifests_for_path(self, path, recursive=False):
        """
        Iterate over loaded Manifests that can apply to path.
        If @recursive is True, returns also Manifests for subdirectories
        of @path. Yields a tuple of (manifest_path, dir_path, manifest).

        The function guarantees that the Manifests for subdirectories
        (more specific) will always be returned before the Manifests
        for parent directories. The order is otherwise undefined.
        """
        return sorted(
                self._iter_unordered_manifests_for_path(
                    path, recursive=recursive),
                key=lambda kdv: len(kdv[1]),
                reverse=True)

    def load_manifests_for_path(self, path, recursive=False, verify=True):
        """
        Load all Manifests that may apply to the specified path,
        recursively. If @recursive is True, also loads Manifests
        for all subdirectories of @path.

        If @verify is True, sub-Manifests will be tested against entries
        in parent Manifests and ManifestMismatch will be raised
        on mismatch. Otherwise, sub-Manifests will be loaded
        unconditionally of whether they match parent checksums.
        """

        pool = multiprocessing.Pool(processes=self.max_jobs)

        try:
            # TODO: figure out how to avoid confusing uses of 'recursive'
            while True:
                to_load = []
                for curmpath, relpath, m in self._iter_manifests_for_path(
                                                path, recursive):
                    for e in m.entries:
                        if e.tag != 'MANIFEST':
                            continue
                        mpath = os.path.join(relpath, e.path)
                        if curmpath == mpath or mpath in self.loaded_manifests:
                            continue
                        mdir = os.path.dirname(mpath)
                        if not verify:
                            e = None
                        if gemato.util.path_starts_with(path, mdir):
                            to_load.append((mpath, e))
                        elif recursive and gemato.util.path_starts_with(mdir, path):
                            to_load.append((mpath, e))
                if not to_load:
                    break

                manifests = pool.imap_unordered(self.manifest_loader, to_load,
                                     chunksize=16)
                self.loaded_manifests.update(manifests)

            pool.close()
            pool.join()
        finally:
            pool.terminate()

    def find_timestamp(self):
        """
        Find a timestamp entry and return it. Returns None if there
        is no timestamp.
        """

        self.load_manifests_for_path('')
        for mpath, p, m in self._iter_manifests_for_path(''):
            for e in m.entries:
                if e.tag == 'TIMESTAMP':
                    return e
        return None

    def set_timestamp(self, ts):
        """
        Set Manifest timestamp to @ts.

        If the Manifest already contains a TIMESTAMP entry, it will
        be updated. Otherwise, a new entry will be created.
        """

        e = self.find_timestamp()
        if e is not None:
            e.ts = ts
        else:
            m = self.loaded_manifests[self.top_level_manifest_filename]
            e = gemato.manifest.ManifestEntryTIMESTAMP(ts)
            m.entries.append(e)

    def find_path_entry(self, path):
        """
        Find a matching entry for path @path and return it. Returns
        None when no path matches. DIST entries are not included.
        """

        self.load_manifests_for_path(path)
        for mpath, relpath, m in self._iter_manifests_for_path(path):
            for e in m.entries:
                if e.tag == 'IGNORE':
                    # ignore matches recursively, so we process it separately
                    # py<3.5 does not have os.path.commonpath()
                    fullpath = os.path.join(relpath, e.path)
                    if gemato.util.path_starts_with(path, fullpath):
                        return e
                elif e.tag in ('DIST', 'TIMESTAMP'):
                    # distfiles are not local files, so skip them
                    # timestamp is not a file ;-)
                    pass
                else:
                    fullpath = os.path.join(relpath, e.path)
                    if fullpath == path:
                        return e
        return None

    def verify_path(self, relpath):
        """
        Verify the path @relpath against appropriate Manifest entry.
        If there is no matching entry, verification fails (as a stray
        file). Returns result as verify_path().
        """
        real_path = os.path.join(self.root_directory, relpath)
        path_entry = self.find_path_entry(relpath)
        return gemato.verify.verify_path(real_path, path_entry)

    def assert_path_verifies(self, relpath):
        """
        Verify the path @relpath against appropriate Manifest entry.
        If there is no matching entry, verification fails (as a stray
        file). Raises exception for failed verification.
        """
        real_path = os.path.join(self.root_directory, relpath)
        path_entry = self.find_path_entry(relpath)
        ret, diff = gemato.verify.verify_path(real_path, path_entry,
                expected_dev=self.manifest_device)
        if not ret:
            raise gemato.exceptions.ManifestMismatch(
                    relpath, path_entry, diff)

    def find_dist_entry(self, filename, relpath=''):
        """
        Find a matching entry for distfile @filename and return it.
        If @relpath is provided, loads all Manifests up to @relpath
        (which can be e.g. a relevant package directory).
        Returns None when no DIST entry matches.
        """

        self.load_manifests_for_path(relpath+'/')
        for mpath, p, m in self._iter_manifests_for_path(relpath+'/'):
            for e in m.entries:
                if e.tag == 'DIST' and e.path == filename:
                    return e
        return None

    def get_file_entry_dict(self, path='', only_types=None,
                            verify_manifests=True):
        """
        Find all file entries that apply to paths starting with @path.
        Returns a nested dictionary that maps directories -> filenames
        inside the directory -> entries. Raises an exception if multiple
        entries for file collide.

        If @only_types are specified as a list, only files of specified
        types will be collected. If it is not specified, then all types
        for local files will be processed.

        @verify_manifests determines whether loaded Manifests will
        be verified against MANIFEST entries. Pass False only when
        doing updates.
        """

        self.load_manifests_for_path(path, recursive=True,
                                     verify=verify_manifests)
        out = {}
        for mpath, relpath, m in self._iter_manifests_for_path(path,
                                    recursive=True):
            for e in m.entries:
                if only_types is not None:
                    if e.tag not in only_types:
                        continue
                    # DIST entries always specify plain filename
                    if e.tag == 'DIST':
                        relpath = ''
                elif e.tag in ('DIST', 'TIMESTAMP'):
                    # distfiles are not local files, so skip them
                    # timestamp is not a file ;-)
                    continue

                fullpath = os.path.join(relpath, e.path)
                if gemato.util.path_starts_with(fullpath, path):
                    dirpath = os.path.dirname(fullpath)
                    filename = os.path.basename(e.path)
                    dirout = out.setdefault(dirpath, {})
                    if filename in dirout:
                        # compare the two entries
                        ret, diff = gemato.verify.verify_entry_compatibility(
                                dirout[filename], e)
                        if not ret:
                            raise gemato.exceptions.ManifestIncompatibleEntry(
                                    dirout[filename], e, diff)
                        # we need to construct a single entry with both checksums
                        if diff:
                            new_checksums = dict(e.checksums)
                            for k, d1, d2 in diff:
                                if d2 is None:
                                    new_checksums[k] = d1
                            e = type(e)(e.path, e.size, new_checksums)
                    dirout[filename] = e
        return out

    def assert_directory_verifies(self, path='',
            fail_handler=gemato.util.throw_exception,
            last_mtime=None):
        """
        Verify the complete directory tree starting at @path (relative
        to top Manifest directory). Includes testing for stray files.
        Raises an exception if any of the files does not pass
        verification.

        @fail_handler is the callback called whenever verification
        fails (ether for mismatch, missing or stray file). The handler
        is passed a ManifestMismatch exception object. The default fail
        handler raises the exception. However, a custom handler can be
        used to provide a non-strict mode, or continue the scan after
        the first failure.

        If none of the handler calls raise an exception, the function
        returns boolean. It returns False if at least one of the handler
        calls returned explicit False; True otherwise.

        If @last_mtime is not None, then only files whose mtime is newer
        than that value (in st_mtime format) will be checked. Use this
        option *only* if mtimes can not be manipulated (i.e. do not use
        it with 'rsync --times')!

        @jobs specifies the number of parallel jobs to use. If set
        to None (the default), the number of system CPUs will be used.
        """

        entry_dict = self.get_file_entry_dict(path)
        it = os.walk(os.path.join(self.root_directory, path),
                onerror=gemato.util.throw_exception,
                followlinks=True)

        def _walk_directory(it):
            """
            Pre-process os.walk() result for verification. Yield objects
            suitable to passing to subprocesses.
            """
            for dirpath, dirnames, filenames in it:
                relpath = os.path.relpath(dirpath, self.root_directory)
                # strip dot to avoid matching problems
                if relpath == '.':
                    relpath = ''
                dirdict = entry_dict.pop(relpath, {})

                skip_dirs = []
                for d in dirnames:
                    # skip dotfiles
                    if d.startswith('.'):
                        skip_dirs.append(d)
                        continue

                    de = dirdict.get(d)
                    if de is None:
                        syspath = os.path.join(dirpath, d)
                        st = os.stat(syspath)
                        if st.st_dev != self.manifest_device:
                            raise gemato.exceptions.ManifestCrossDevice(syspath)
                        continue

                    if de.tag == 'IGNORE':
                        skip_dirs.append(d)
                        del dirdict[d]

                # skip scanning ignored directories
                for d in skip_dirs:
                    dirnames.remove(d)

                yield (dirpath, relpath, dirnames, filenames, dirdict)

        verifier = SubprocessVerifier(
                self.top_level_manifest_filename,
                self.manifest_device,
                fail_handler, last_mtime)

        pool = multiprocessing.Pool(processes=self.max_jobs)

        try:
            # verify the directories in parallel
            if sys.hexversion >= 0x03050400:
                ret = all(pool.imap_unordered(verifier, _walk_directory(it),
                                   chunksize=64))
            else:
                # in py<3.5.4 imap() swallows exceptions, so fall back
                # to regular map() [it's only a little slower]
                ret = all(pool.map(verifier, _walk_directory(it),
                                   chunksize=64))

            pool.close()

            # check for missing directories
            for relpath, dirdict in entry_dict.items():
                for f, e in dirdict.items():
                    fpath = os.path.join(relpath, f)
                    syspath = os.path.join(self.root_directory, fpath)
                    ret &= verifier._verify_one_file(syspath, fpath, e)

            pool.join()
        finally:
            pool.terminate()

        return ret

    def save_manifests(self, hashes=None, force=False, sort=None,
            compress_watermark=None, compress_format=None):
        """
        Save the Manifests modified since the last save_manifests()
        call.

        @hashes, @sort, @compress_watermark and @compress_format
        override the value specified in the constructor. If None,
        the values from the constructor are used. If those were None
        as well, the defaults are used.

        @hashes specifies the requested hash set. The effective value
        must be non-null since new entries can be created.

        If @force is True, all Manifests will be rewritten even
        if they were not modified.

        If @sort is True, the Manifest entries will be sorted prior
        to saving. By default they are not.

        If @compress_watermark is not None, then the uncompressed
        Manifest files whose size is larger than or equal to the value
        will be compressed using @compress_format. The Manifest files
        whose size is smaller will be uncompressed. To compress all
        Manifest files, pass a size of 0.
        
        If @compress_watermark is None, the compression is left as-is.
        """

        if hashes is None:
            hashes = self.hashes
        if sort is None:
            sort = self.sort
        if compress_watermark is None:
            compress_watermark = self.compress_watermark
        if compress_format is None:
            compress_format = self.compress_format
        if force:
            self.load_manifests_for_path('', recursive=True)

        fixed_manifests = set()
        renamed_manifests = {}
        for mpath, relpath, m in self._iter_manifests_for_path('',
                                    recursive=True):
            for e in m.entries:
                if e.tag != 'MANIFEST':
                    continue

                fullpath = os.path.join(relpath, e.path)
                if not force and fullpath not in self.updated_manifests:
                    assert fullpath not in renamed_manifests
                    continue
                if fullpath in renamed_manifests:
                    fullpath = renamed_manifests[fullpath]
                    e.path = os.path.relpath(fullpath, relpath)

                gemato.verify.update_entry_for_path(
                    os.path.join(self.root_directory, fullpath),
                    e,
                    hashes=hashes,
                    expected_dev=self.manifest_device)

                # do not remove it from self.updated_manifests
                # immediately as we may have to deal with multiple
                # entries
                fixed_manifests.add(fullpath)
                self.updated_manifests.add(mpath)

            # we've apparently modified this Manifest, so store it now
            if force or mpath in self.updated_manifests:
                unc_size = self.save_manifest(mpath, sort=sort)
                # let's see if we want to recompress it
                if compress_watermark is not None:
                    compr = (gemato.compression
                            .get_compressed_suffix_from_filename(mpath))
                    is_compr = compr is not None
                    want_compr = self.profile.want_compressed_manifest(
                            mpath, m, unc_size, compress_watermark)
                    if want_compr is not None and is_compr != want_compr:
                        if want_compr:
                            # compress it!
                            new_mpath = mpath + '.' + compress_format
                        else:
                            new_mpath = mpath[:-len(compr)-1]

                        # do the rename!
                        self.loaded_manifests[new_mpath] = m
                        self.save_manifest(new_mpath)
                        del self.loaded_manifests[mpath]
                        os.unlink(os.path.join(self.root_directory,
                            mpath))
                        renamed_manifests[mpath] = new_mpath

                        if mpath == self.top_level_manifest_filename:
                            self.top_level_manifest_filename = new_mpath

        # now, discard all the Manifests whose entries we've updated
        self.updated_manifests -= fixed_manifests
        # ...and those which we renamed
        self.updated_manifests -= set(renamed_manifests.keys())
        # ...and top-level Manifest which has no entries
        self.updated_manifests.discard(self.top_level_manifest_filename)
        # at this point, the list should be empty
        assert not self.updated_manifests, (
                "Unlinked but updated Manifests: {}".format(
                    self.updated_manifests))

    def update_entry_for_path(self, path, new_entry_type='DATA',
            hashes=None):
        """
        Update the Manifest entries for @path and queue the containing
        Manifests for update. @path must not be covered by IGNORE.
        You need to invoke save_manifests() to store the Manifest
        updates afterwards.

        If the path exists and has a matching Manifest entry, the most
        specific existing entry will be updated. If the path has more
        entries, the remaining entries will be removed. This function
        does not check if they were compatible.

        The type of MANIFEST and DATA derived entries is preserved.

        If the path exists and has no Manifest entry, a new entry
        of type @new_entry_type will be created in the Manifest most
        specific to the location. Note that AUX entries can only
        be created if they're located in 'files/' directory relative
        to an existing Manifest.

        If the path does not exist, all Manifest entries for it will
        be removed.

        @hashes override the value specified in the constructor.
        If None, the values from the constructor are used. If those were
        None as well, the defaults are used.

        @hashes specifies the requested hash set. If the effective value
        is None, the routine reuses the existing hash set in the entry.
        When creating a new entry, @hashes must be non-null.
        """

        had_entry = False
        if hashes is None:
            hashes = self.hashes

        self.load_manifests_for_path(path)
        for mpath, relpath, m in self._iter_manifests_for_path(path):
            entries_to_remove = []
            for e in m.entries:
                if e.tag == 'IGNORE':
                    # ignore matches recursively, so we process it separately
                    # py<3.5 does not have os.path.commonpath()
                    fullpath = os.path.join(relpath, e.path)
                    assert not gemato.util.path_starts_with(path, fullpath)
                elif e.tag in ('DIST', 'TIMESTAMP'):
                    # distfiles are not local files, so skip them
                    # timestamp is not a file ;-)
                    pass
                else:
                    # we update either file at the specified path
                    # or any relevant Manifests
                    fullpath = os.path.join(relpath, e.path)
                    if fullpath != path:
                        continue

                    if had_entry:
                        # duplicate entry!
                        entries_to_remove.append(e)
                        continue

                    try:
                        gemato.verify.update_entry_for_path(
                            os.path.join(self.root_directory,
                                fullpath),
                            e,
                            hashes=hashes,
                            expected_dev=self.manifest_device)
                    except gemato.exceptions.ManifestInvalidPath as err:
                        if err.detail[0] == '__exists__':
                            # file does not exist anymore, so remove
                            # the entry
                            entries_to_remove.append(e)
                            had_entry = True
                        else:
                            raise err
                    else:
                        self.updated_manifests.add(mpath)
                        had_entry = True

            if entries_to_remove:
                for e in entries_to_remove:
                    m.entries.remove(e)
                self.updated_manifests.add(mpath)

        if not had_entry:
            assert hashes is not None
            for mpath, relpath, m in self._iter_manifests_for_path(path):
                # add to the first relevant Manifest
                assert new_entry_type not in ('DIST', 'IGNORE')
                newpath = os.path.relpath(path, relpath)
                if new_entry_type == 'AUX':
                    # AUX has implicit files/ prefix
                    assert gemato.util.path_inside_dir(newpath,
                            'files')
                    # drop files/ prefix
                    newpath = os.path.relpath(newpath, 'files')
                e = gemato.manifest.new_manifest_entry(
                        new_entry_type, newpath, 0, {})
                gemato.verify.update_entry_for_path(
                    os.path.join(self.root_directory, path),
                    e,
                    hashes=hashes,
                    expected_dev=self.manifest_device)
                m.entries.append(e)
                self.updated_manifests.add(mpath)
                had_entry = True
                break

    def get_deduplicated_file_entry_dict_for_update(self, path='',
            verify_manifests=True):
        """
        Find all file entries that apply to paths starting with @path.
        Remove all duplicate entries and queue the relevant Manifests
        for update. Return a dictionary mapping relative paths
        to tuple of (manifest path, entry).

        You need to invoke save_manifests() to store the Manifest
        updates afterwards. However, note that the resulting tree
        may no longer validate.

        If the path is referenced by multiple entries of incompatible
        semantics, raises an exception. If the entries have compatible
        semantics, all but the first (deepest) are removed, even
        if they have colliding sizes or hashes. If the duplicate
        entries use different hash sets, the preserved entry is updated
        to have the union of their hashes.

        @verify_manifests determines whether loaded Manifests will
        be verified against MANIFEST entries. Pass False only when
        doing updates.
        """

        self.load_manifests_for_path(path, recursive=True,
                                     verify=verify_manifests)
        out = {}
        for mpath, relpath, m in self._iter_manifests_for_path(path,
                                    recursive=True):
            entries_to_remove = []
            for e in m.entries:
                if e.tag in ('DIST', 'TIMESTAMP'):
                    # distfiles are not local files, so skip them
                    # timestamp is not a file ;-)
                    continue

                fullpath = os.path.join(relpath, e.path)
                if gemato.util.path_starts_with(fullpath, path):
                    if fullpath in out:
                        # compare the two entries
                        ret, diff = gemato.verify.verify_entry_compatibility(
                                out[fullpath][1], e)
                        # if semantically incompatible, throw
                        if not ret and diff[0][0] == '__type__':
                            raise (gemato.exceptions
                                    .ManifestIncompatibleEntry(
                                        out[fullpath][1], e, diff))
                        # otherwise, make sure we have all checksums
                        out[fullpath][1].checksums.update(e.checksums)
                        # and drop the duplicate
                        entries_to_remove.append(e)
                    else:
                        out[fullpath] = (mpath, e)

            if entries_to_remove:
                for e in entries_to_remove:
                    m.entries.remove(e)
                self.updated_manifests.add(mpath)

        return out

    def load_unregistered_manifests(self, path='', verify_manifests=True):
        """
        Scan the directory @path (relative to top directory)
        for unregistered (not listed in MANIFEST entries) Manifest
        files and load them if they are valid.

        Returns a list of files found. The respective MANIFEST entries
        need to be added to other Manifests manually to ensure
        integrity. Note that the list may contain files that are
        referenced within added Manifests, so the list should
        be verified with regards to existing entries.

        @verify_manifests determines whether loaded Manifests will
        be verified against MANIFEST entries. Pass False only when
        doing updates.
        """

        manifest_filenames = (gemato.compression
                .get_potential_compressed_names('Manifest'))

        entry_dict = self.get_file_entry_dict(path,
                only_types=['IGNORE'], verify_manifests=verify_manifests)
        new_manifests = []
        it = os.walk(os.path.join(self.root_directory, path),
                onerror=gemato.util.throw_exception,
                followlinks=True)

        for dirpath, dirnames, filenames in it:
            relpath = os.path.relpath(dirpath, self.root_directory)
            # strip dot to avoid matching problems
            if relpath == '.':
                relpath = ''
            dirdict = entry_dict.get(relpath, {})

            skip_dirs = []
            for d in dirnames:
                # skip dotfiles
                if d.startswith('.'):
                    skip_dirs.append(d)
                    continue

                de = dirdict.get(d, None)
                if de is None:
                    syspath = os.path.join(dirpath, d)
                    st = os.stat(syspath)
                    if st.st_dev != self.manifest_device:
                        raise gemato.exceptions.ManifestCrossDevice(syspath)
                    continue

                assert de.tag == 'IGNORE'
                skip_dirs.append(d)

            # skip scanning ignored directories
            for d in skip_dirs:
                dirnames.remove(d)

            # check for unregistered Manifest
            for mname in manifest_filenames:
                if mname in filenames:
                    fpath = os.path.join(relpath, mname)
                    if fpath in self.loaded_manifests:
                        continue

                    # we've just found ourselves a new Manifest,
                    # let's try to load it
                    try:
                        self.load_manifest(fpath)
                    except gemato.exceptions.ManifestSyntaxError:
                        # syntax error? probably not a Manifest then.
                        pass
                    else:
                        new_manifests.append(fpath)

        return new_manifests


    def update_entries_for_directory(self, path='', hashes=None,
            last_mtime=None, verify_manifests=False):
        """
        Update the Manifest entries for the contents of directory
        @path (top directory by default), recursively. Includes adding
        new files and removing entries for those that no longer exist.
        The behavior for various cases is the same
        as for update_entry_for_path() except as noted below.

        New entries are currently created with DATA type. This will
        be extended in the future.

        @hashes override the value specified in the constructor.
        If None, the values from the constructor are used. If those were
        None as well, the defaults are used.

        @hashes specifies the requested hash set. The effective value
        must be non-null since new entries can be created.

        If @last_mtime is not None, then only files whose mtime is newer
        than that value (in st_mtime format) will be updated. Use this
        option *only* if you can rely on mtimes being bumped
        monotonically on modified files. Afterwards, the value
        of @last_mtime should be put into the TIMESTAMP entry.

        @verify_manifests determines whether loaded Manifests will
        be verified against MANIFEST entries. Disabled by default since
        the MANIFEST entries would be updated anyway.
        """

        if hashes is None:
            hashes = self.hashes
        assert hashes is not None

        manifest_filenames = (gemato.compression
                .get_potential_compressed_names('Manifest'))

        new_manifests = self.load_unregistered_manifests(path,
                verify_manifests=verify_manifests)
        entry_dict = self.get_deduplicated_file_entry_dict_for_update(
                path, verify_manifests=verify_manifests)
        manifest_stack = []
        for mpath, mrpath, m in (self
                ._iter_manifests_for_path(path)):
            manifest_stack.append((mpath, mrpath, m))
            break

        it = os.walk(os.path.join(self.root_directory, path),
                onerror=gemato.util.throw_exception,
                followlinks=True)

        for dirpath, dirnames, filenames in it:
            relpath = os.path.relpath(dirpath, self.root_directory)
            # strip dot to avoid matching problems
            if relpath == '.':
                relpath = ''

            # drop Manifest paths until we get to a common directory
            while not gemato.util.path_starts_with(relpath,
                    manifest_stack[-1][1]):
                manifest_stack.pop()

            want_manifest = self.profile.want_manifest_in_directory(
                    relpath, dirnames, filenames)

            skip_dirs = []
            for d in dirnames:
                # skip dotfiles
                if d.startswith('.'):
                    skip_dirs.append(d)
                    continue

                dpath = os.path.join(relpath, d)
                mpath, de = entry_dict.pop(dpath, (None, None))
                if de is None:
                    syspath = os.path.join(dirpath, d)
                    st = os.stat(syspath)
                    if st.st_dev != self.manifest_device:
                        raise gemato.exceptions.ManifestCrossDevice(syspath)
                    continue

                if de.tag == 'IGNORE':
                    skip_dirs.append(d)
                else:
                    # trigger the exception indirectly
                    gemato.verify.update_entry_for_path(
                        os.path.join(dirpath, d),
                        de,
                        hashes=hashes,
                        expected_dev=self.manifest_device)
                    assert False, "exception should have been raised"

            # skip scanning ignored directories
            for d in skip_dirs:
                dirnames.remove(d)

            new_entries = []
            for f in filenames:
                # skip dotfiles
                if f.startswith('.'):
                    continue

                fpath = os.path.join(relpath, f)
                mpath, fe = entry_dict.pop(fpath, (None, None))
                if fe is not None:
                    if fe.tag == 'IGNORE':
                        continue
                    if fe.tag == 'MANIFEST':
                        manifest_stack.append((fpath, relpath,
                            self.loaded_manifests[fpath]))
                        # do not update the Manifest entry if
                        # the relevant Manifest is going to be updated
                        # anyway
                        if relpath in self.updated_manifests:
                            continue
                else:
                    # skip top-level Manifest, we obviously can't have
                    # an entry for it
                    if fpath in manifest_filenames:
                        continue
                    if fpath in new_manifests:
                        ftype = 'MANIFEST'
                        manifest_stack.append((fpath, relpath,
                            self.loaded_manifests[fpath]))
                    else:
                        ftype = self.profile.get_entry_type_for_path(
                                fpath)

                    # note: .path needs to be corrected below
                    fe = gemato.manifest.new_manifest_entry(ftype,
                            fpath, 0, {})
                    new_entries.append(fe)
                    if relpath in self.updated_manifests:
                        continue

                changed = gemato.verify.update_entry_for_path(
                    os.path.join(dirpath, f),
                    fe,
                    hashes=hashes,
                    expected_dev=self.manifest_device,
                    last_mtime=last_mtime)
                if changed and mpath is not None:
                    self.updated_manifests.add(mpath)

            # do we have Manifest in this directory?
            new_ignore_paths = []
            if want_manifest and manifest_stack[-1][1] != relpath:
                mpath = os.path.join(relpath, 'Manifest')
                m = self.create_manifest(mpath)
                manifest_stack.append((mpath, relpath, m))
                fe = gemato.manifest.ManifestEntryMANIFEST(
                        mpath, 0, {})
                new_entries.append(fe)

                for ip in (self.profile
                           .get_ignore_paths_for_new_manifest(relpath)):
                    ie = gemato.manifest.ManifestEntryIGNORE(ip)
                    iep = os.path.join(relpath, ip)

                    if self.find_path_entry(iep):
                        raise NotImplementedError('Need to remove old parent entry for now-ignored path')

                    m.entries.append(ie)
                    new_ignore_paths.append(iep)

            if new_entries:
                mpath, mdirpath, m = manifest_stack[-1]
                for fe in new_entries:
                    # skip files that should have been ignored
                    if fe.path in new_ignore_paths:
                        continue

                    if fe.tag == 'MANIFEST':
                        # Manifest needs to go level up
                        mmpath = mpath
                        mm = m
                        mmdirpath = mdirpath
                        i = -1
                        while mmdirpath == os.path.dirname(fe.path):
                            i -= 1
                            mmpath, mmdirpath, mm = manifest_stack[i]

                        fe.path = os.path.relpath(fe.path, mmdirpath)
                        mm.entries.append(fe)
                        self.updated_manifests.add(mmpath)
                    else:
                        if ftype == 'AUX':
                            # AUX has implicit files/ prefix in .path
                            # but for now, we've shoved our path
                            # into .aux_path
                            fe.path = os.path.relpath(fe.aux_path,
                                    mdirpath)
                            assert gemato.util.path_inside_dir(
                                    fe.path, 'files')
                            # drop files/ prefix for the entry
                            fe.aux_path = os.path.relpath(fe.path,
                                    'files')
                        else:
                            fe.path = os.path.relpath(fe.path, mdirpath)
                        # do not add duplicate entry if the path is ignored
                        m.entries.append(fe)
                self.updated_manifests.add(mpath)

        # check for removed files
        for relpath, me in entry_dict.items():
            mpath, fe = me
            if fe.tag == 'IGNORE':
                continue

            self.loaded_manifests[mpath].entries.remove(fe)
            self.updated_manifests.add(mpath)

    def create_manifest(self, path):
        """
        Create a new empty sub-Manifest instance at relative path @path.
        The file will not be written until save_manifests(). No MANIFEST
        entry for the file will be created.

        Returns the new ManifestFile instance.
        """

        return self.load_manifest(path, allow_create=True)



# gemato: CLI routines
# vim:fileencoding=utf-8
# (c) 2017-2018 Michał Górny
# Licensed under the terms of 2-clause BSD license

from __future__ import print_function

import argparse
import datetime
import io
import logging
import multiprocessing
import os.path
import sys
import timeit

import gemato.find_top_level
import gemato.profile
import gemato.recursiveloader


def verify_failure(e):
    logging.error(str(e))
    return False


def do_verify(args, argp):
    ret = True

    for p in args.paths:
        tlm = gemato.find_top_level.find_top_level_manifest(p)
        if tlm is None:
            logging.error('Top-level Manifest not found in {}'.format(p))
            return 1

        init_kwargs = {}
        kwargs = {}
        if args.jobs is not None:
            if args.jobs < 1:
                argp.error('--jobs must be positive')
            init_kwargs['max_jobs'] = args.jobs
        if args.keep_going:
            kwargs['fail_handler'] = verify_failure
        if not args.openpgp_verify:
            init_kwargs['verify_openpgp'] = False

        # use isolated environment if key is specified;
        # system environment otherwise
        if args.openpgp_key is not None:
            env_class = gemato.openpgp.OpenPGPEnvironment
        else:
            env_class = gemato.openpgp.OpenPGPSystemEnvironment

        with env_class() as env:
            if args.openpgp_key is not None:
                with io.open(args.openpgp_key, 'rb') as f:
                    env.import_key(f)
                # always refresh keys to check for revocation
                # (unless user specifically asked us not to)
                if args.refresh_keys:
                    logging.info('Refreshing keys from keyserver...')
                    env.refresh_keys()
                    logging.info('Keys refreshed.')
            init_kwargs['openpgp_env'] = env

            start = timeit.default_timer()
            try:
                m = gemato.recursiveloader.ManifestRecursiveLoader(tlm, **init_kwargs)
            except gemato.exceptions.OpenPGPNoImplementation as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.OpenPGPVerificationFailure as e:
                logging.error(str(e))
                return 1
            if args.require_signed_manifest and not m.openpgp_signed:
                logging.error('Top-level Manifest {} is not OpenPGP signed'.format(tlm))
                return 1

            ts = m.find_timestamp()
            if ts:
                logging.info('Manifest timestamp: {} UTC'.format(ts.ts))

            if m.openpgp_signed:
                logging.info('Valid OpenPGP signature found:')
                logging.info('- primary key: {}'.format(
                    m.openpgp_signature.primary_key_fingerprint))
                logging.info('- subkey: {}'.format(
                    m.openpgp_signature.fingerprint))
                logging.info('- timestamp: {} UTC'.format(
                    m.openpgp_signature.timestamp))

            logging.info('Verifying {}...'.format(p))

            relpath = os.path.relpath(p, os.path.dirname(tlm))
            if relpath == '.':
                relpath = ''
            try:
                ret &= m.assert_directory_verifies(relpath, **kwargs)
            except gemato.exceptions.ManifestCrossDevice as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.ManifestIncompatibleEntry as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.ManifestMismatch as e:
                logging.error(str(e))
                return 1

            stop = timeit.default_timer()
            logging.info('{} verified in {:.2f} seconds'.format(p, stop - start))
    return 0 if ret else 1


def do_update(args, argp):
    for p in args.paths:
        tlm = gemato.find_top_level.find_top_level_manifest(p)
        if tlm is None:
            logging.error('Top-level Manifest not found in {}'.format(p))
            return 1

        init_kwargs = {}
        save_kwargs = {}
        update_kwargs = {}
        if args.hashes is not None:
            init_kwargs['hashes'] = args.hashes.split()
        if args.compress_watermark is not None:
            if args.compress_watermark < 0:
                argp.error('--compress-watermark must not be negative!')
            init_kwargs['compress_watermark'] = args.compress_watermark
        if args.compress_format is not None:
            init_kwargs['compress_format'] = args.compress_format
        if args.force_rewrite:
            save_kwargs['force'] = True
        if args.jobs is not None:
            if args.jobs < 1:
                argp.error('--jobs must be positive')
            init_kwargs['max_jobs'] = args.jobs
        if args.openpgp_id is not None:
            init_kwargs['openpgp_keyid'] = args.openpgp_id
        if args.profile is not None:
            init_kwargs['profile'] = gemato.profile.get_profile_by_name(
                    args.profile)
        if args.sign is not None:
            init_kwargs['sign_openpgp'] = args.sign

        # use isolated environment if key is specified;
        # system environment otherwise
        if args.openpgp_key is not None:
            env_class = gemato.openpgp.OpenPGPEnvironment
        else:
            env_class = gemato.openpgp.OpenPGPSystemEnvironment

        with env_class() as env:
            if args.openpgp_key is not None:
                with io.open(args.openpgp_key, 'rb') as f:
                    env.import_key(f)
            init_kwargs['openpgp_env'] = env

            start = timeit.default_timer()
            try:
                m = gemato.recursiveloader.ManifestRecursiveLoader(tlm,
                        **init_kwargs)
            except gemato.exceptions.OpenPGPNoImplementation as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.OpenPGPVerificationFailure as e:
                logging.error(str(e))
                return 1

            # if not specified by user, profile must set it
            if m.hashes is None:
                argp.error('--hashes must be specified if not implied by --profile')

            relpath = os.path.relpath(p, os.path.dirname(tlm))
            if relpath == '.':
                relpath = ''
            if args.timestamp and relpath != '':
                argp.error('Timestamp can only be updated if doing full-tree update')
            if args.incremental:
                if relpath != '':
                    argp.error('Incremental works only for full-tree update')
                last_ts = m.find_timestamp()
                if last_ts is None:
                    argp.error('Incremental specified but no timestamp in Manifest')
                update_kwargs['last_mtime'] = last_ts.ts.timestamp()

            logging.info('Updating Manifests in {}...'.format(p))

            try:
                start_ts = datetime.datetime.utcnow()
                m.update_entries_for_directory(relpath, **update_kwargs)

                # write TIMESTAMP if requested, or if already there
                if relpath != '':
                    # skip timestamp if not doing full update
                    pass
                elif args.timestamp:
                    m.set_timestamp(start_ts)
                else:
                    ts = m.find_timestamp()
                    if ts is not None:
                        ts.ts = start_ts

                m.save_manifests(**save_kwargs)
            except gemato.exceptions.ManifestCrossDevice as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.ManifestInvalidPath as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.ManifestInvalidFilename as e:
                logging.error(str(e))
                return 1

            stop = timeit.default_timer()
            logging.info('{} updated in {:.2f} seconds'.format(p, stop - start))
    return 0


def do_create(args, argp):
    for p in args.paths:
        init_kwargs = {}
        save_kwargs = {}
        init_kwargs['allow_create'] = True
        if args.hashes is not None:
            init_kwargs['hashes'] = args.hashes.split()
        if args.compress_watermark is not None:
            if args.compress_watermark < 0:
                argp.error('--compress-watermark must not be negative!')
            init_kwargs['compress_watermark'] = args.compress_watermark
        if args.compress_format is not None:
            init_kwargs['compress_format'] = args.compress_format
        if args.force_rewrite:
            save_kwargs['force'] = True
        if args.jobs is not None:
            if args.jobs < 1:
                argp.error('--jobs must be positive')
            init_kwargs['max_jobs'] = args.jobs
        if args.openpgp_id is not None:
            init_kwargs['openpgp_keyid'] = args.openpgp_id
        if args.profile is not None:
            init_kwargs['profile'] = gemato.profile.get_profile_by_name(
                    args.profile)
        if args.sign is not None:
            init_kwargs['sign_openpgp'] = args.sign

        # use isolated environment if key is specified;
        # system environment otherwise
        if args.openpgp_key is not None:
            env_class = gemato.openpgp.OpenPGPEnvironment
        else:
            env_class = gemato.openpgp.OpenPGPSystemEnvironment

        with env_class() as env:
            if args.openpgp_key is not None:
                with io.open(args.openpgp_key, 'rb') as f:
                    env.import_key(f)
            init_kwargs['openpgp_env'] = env

            start = timeit.default_timer()
            try:
                m = gemato.recursiveloader.ManifestRecursiveLoader(
                        os.path.join(p, 'Manifest'), **init_kwargs)
            except gemato.exceptions.OpenPGPNoImplementation as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.OpenPGPVerificationFailure as e:
                logging.error(str(e))
                return 1

            # if not specified by user, profile must set it
            if m.hashes is None:
                argp.error('--hashes must be specified if not implied by --profile')

            logging.info('Creating Manifests in {}...'.format(p))

            try:
                start_ts = datetime.datetime.utcnow()
                m.update_entries_for_directory()

                # write TIMESTAMP if requested, or if already there
                if args.timestamp:
                    m.set_timestamp(start_ts)

                m.save_manifests(**save_kwargs)
            except gemato.exceptions.ManifestCrossDevice as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.ManifestInvalidPath as e:
                logging.error(str(e))
                return 1
            except gemato.exceptions.ManifestInvalidFilename as e:
                logging.error(str(e))
                return 1

            stop = timeit.default_timer()
            logging.info('{} updated in {:.2f} seconds'.format(p, stop - start))
    return 0


def main(argv):
    argp = argparse.ArgumentParser(
            prog=argv[0],
            description='Gentoo Manifest Tool')
    subp = argp.add_subparsers()

    verify = subp.add_parser('verify',
            help='Verify one or more directories against Manifests')
    verify.add_argument('paths', nargs='*', default=['.'],
            help='Paths to verify (defaults to "." if none specified)')
    verify.add_argument('-j', '--jobs', type=int,
            help='Specify the maximum number of parallel jobs to use (default: {})'
                .format(multiprocessing.cpu_count()))
    verify.add_argument('-k', '--keep-going', action='store_true',
            help='Continue reporting errors rather than terminating on the first failure')
    verify.add_argument('-K', '--openpgp-key',
            help='Use only the OpenPGP key(s) from a specific file')
    verify.add_argument('-P', '--no-openpgp-verify', action='store_false',
            dest='openpgp_verify',
            help='Disable OpenPGP verification of signed Manifests')
    verify.add_argument('-R', '--no-refresh-keys', action='store_false',
            dest='refresh_keys',
            help='Disable refreshing OpenPGP key (prevents network access, applicable '
                +'when using -K only)')
    verify.add_argument('-s', '--require-signed-manifest', action='store_true',
            help='Require that the top-level Manifest is OpenPGP signed')
    verify.set_defaults(func=do_verify)

    update = subp.add_parser('update',
            help='Update the Manifest entries for one or more directory trees')
    update.add_argument('paths', nargs='*', default=['.'],
            help='Paths to update (defaults to "." if none specified)')
    update.add_argument('-c', '--compress-watermark', type=int,
            help='Minimum Manifest size for files to be compressed')
    update.add_argument('-C', '--compress-format',
            help='Format for compressed files (e.g. "gz", "bz2"...)')
    update.add_argument('-f', '--force-rewrite', action='store_true',
            help='Force rewriting all the Manifests, even if they did not change')
    update.add_argument('-H', '--hashes',
            help='Whitespace-separated list of hashes to use')
    update.add_argument('-i', '--incremental', action='store_true',
            help='Perform incremental update by comparing mtimes against TIMESTAMP')
    update.add_argument('-j', '--jobs', type=int,
            help='Specify the maximum number of parallel jobs to use (default: {})'
                .format(multiprocessing.cpu_count()))
    update.add_argument('-k', '--openpgp-id',
            help='Use the specified OpenPGP key (by ID or user)')
    update.add_argument('-K', '--openpgp-key',
            help='Use only the OpenPGP key(s) from a specific file')
    update.add_argument('-p', '--profile',
            help='Use the specified profile ("default", "ebuild", "old-ebuild"...)')
    signgroup = update.add_mutually_exclusive_group()
    signgroup.add_argument('-s', '--sign', action='store_true',
            default=None,
            help='Force signing the top-level Manifest')
    signgroup.add_argument('-S', '--no-sign', action='store_false',
            dest='sign',
            help='Disable signing the top-level Manifest')
    update.add_argument('-t', '--timestamp', action='store_true',
            help='Include TIMESTAMP entry in Manifest')
    update.set_defaults(func=do_update)

    create = subp.add_parser('create',
            help='Create a Manifest tree starting at the specified file')
    create.add_argument('paths', nargs='*', default=['.'],
            help='Paths to create (defaults to "Manifest" if none specified)')
    create.add_argument('-c', '--compress-watermark', type=int,
            help='Minimum Manifest size for files to be compressed')
    create.add_argument('-C', '--compress-format',
            help='Format for compressed files (e.g. "gz", "bz2"...)')
    create.add_argument('-f', '--force-rewrite', action='store_true',
            help='Force rewriting all the Manifests, even if they did not change')
    create.add_argument('-H', '--hashes',
            help='Whitespace-separated list of hashes to use')
    create.add_argument('-j', '--jobs', type=int,
            help='Specify the maximum number of parallel jobs to use (default: {})'
                .format(multiprocessing.cpu_count()))
    create.add_argument('-k', '--openpgp-id',
            help='Use the specified OpenPGP key (by ID or user)')
    create.add_argument('-K', '--openpgp-key',
            help='Use only the OpenPGP key(s) from a specific file')
    create.add_argument('-p', '--profile',
            help='Use the specified profile ("default", "ebuild", "old-ebuild"...)')
    signgroup = create.add_mutually_exclusive_group()
    signgroup.add_argument('-s', '--sign', action='store_true',
            default=None,
            help='Force signing the top-level Manifest')
    signgroup.add_argument('-S', '--no-sign', action='store_false',
            dest='sign',
            help='Disable signing the top-level Manifest')
    create.add_argument('-t', '--timestamp', action='store_true',
            help='Include TIMESTAMP entry in Manifest')
    create.set_defaults(func=do_create)

    vals = argp.parse_args(argv[1:])
    if not hasattr(vals, 'func'):
        argp.error('No function specified')
    return vals.func(vals, argp)


def setuptools_main():
    logging.getLogger().setLevel(logging.INFO)
    sys.exit(main(sys.argv))



# gemato: Top-level Manifest finding tests
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license

import gzip
import os
import os.path
import unittest

import gemato.find_top_level

from tests.testutil import TempDirTestCase


class TestCurrentDirectory(TempDirTestCase):
    """
    Test for finding top-level Manifest in a plain tree.
    """

    DIRS = ['suba', 'subb', 'subc', 'subc/sub']
    FILES = {
        'Manifest': u'',
        'subb/Manifest': u'',
        'subc/sub/Manifest': u'',
    }

    def test_find_top_level_manifest(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(self.dir),
                    self.dir),
                'Manifest')

    def test_find_top_level_manifest_from_empty_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'suba')),
                    self.dir),
                'Manifest')

    def test_find_top_level_manifest_from_manifest_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'subb')),
                    self.dir),
                'Manifest')

    def test_find_top_level_manifest_from_deep_manifest_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'subc', 'sub')),
                    self.dir),
                'Manifest')


class TestUnreadableManifest(TempDirTestCase):
    """
    Test whether the function fails correctly when it can not read
    a Manifest file.
    """

    FILES = {
        'Manifest': u'',
    }

    def setUp(self):
        super(TestUnreadableManifest, self).setUp()
        os.chmod(os.path.join(self.dir, 'Manifest'), 0)

    def test_find_top_level_manifest(self):
        self.assertRaises(IOError,
                gemato.find_top_level.find_top_level_manifest, self.dir)


class TestIgnoredSubdir(TempDirTestCase):
    """
    Test for ignoring irrelevant Manifest.
    """

    DIRS = ['sub', 'sub/sub', 'subb', 'subempty']
    FILES = {
        'Manifest': u'''
IGNORE sub
IGNORE subempty
''',
        'sub/Manifest': u'',
    }

    def test_find_top_level_manifest(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(self.dir),
                    self.dir),
                'Manifest')

    def test_find_top_level_manifest_from_ignored_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'sub')),
                    self.dir),
                'sub/Manifest')

    def test_find_top_level_manifest_from_sub_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'sub/sub')),
                    self.dir),
                'sub/Manifest')

    def test_find_top_level_manifest_from_non_ignored_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'subb')),
                    self.dir),
                'Manifest')

    def test_find_top_level_manifest_from_ignored_empty_subdir(self):
        self.assertIsNone(
                gemato.find_top_level.find_top_level_manifest(
                    os.path.join(self.dir, 'subempty')))


class TestEmptyTree(TempDirTestCase):
    """
    Test for finding top-level Manifest in a tree without a Manifest
    """

    def test_find_top_level_manifest(self):
        self.assertIsNone(
                gemato.find_top_level.find_top_level_manifest(self.dir))


class TestRootDirectory(unittest.TestCase):
    """
    Test behavior when run on the system root directory.
    """

    def test_find_top_level_manifest(self):
        if os.path.exists('/Manifest'):
            raise unittest.SkipTest('/Manifest is present')
        self.assertIsNone(
                gemato.find_top_level.find_top_level_manifest('/'))


class TestCrossDevice(TempDirTestCase):
    """
    Test behavior when attempting to cross device boundary.
    """

    FILES = {
        'Manifest': u'',
    }

    def setUp(self):
        if not os.path.exists('/proc'):
            raise unittest.SkipTest('/proc does not exist')
        super(TestCrossDevice, self).setUp()
        os.symlink('/proc', os.path.join(self.dir, 'test'))

    def test_find_top_level_manifest(self):
        self.assertIsNone(
                gemato.find_top_level.find_top_level_manifest(
                    os.path.join(self.dir, 'test')))


class TestCrossDeviceManifest(TempDirTestCase):
    """
    Test behavior when attempting to use a Manifest from other device
    (symlinked).
    """

    DIRS = ['sub']

    def setUp(self):
        if not os.path.exists('/proc/version'):
            raise unittest.SkipTest('/proc/version does not exist')
        super(TestCrossDeviceManifest, self).setUp()
        os.symlink('/proc/version', os.path.join(self.dir, 'Manifest'))

    def test_find_top_level_manifest(self):
        self.assertIsNone(
                gemato.find_top_level.find_top_level_manifest(
                    os.path.join(self.dir, 'sub')))


class TestCompressedManifest(TempDirTestCase):
    """
    Test for finding compressed Manifest in a plain tree.
    """

    DIRS = ['suba', 'subb', 'subc', 'subc/sub']
    FILES = {
        'subb/Manifest': u'',
    }

    def setUp(self):
        super(TestCompressedManifest, self).setUp()
        with gzip.GzipFile(os.path.join(self.dir, 'Manifest.gz'), 'wb'):
            pass
        with gzip.GzipFile(os.path.join(self.dir, 'subc/sub/Manifest.gz'), 'wb'):
            pass

    def test_find_top_level_manifest_no_allow_compressed(self):
        self.assertIsNone(
                gemato.find_top_level.find_top_level_manifest(self.dir))

    def test_find_top_level_manifest(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        self.dir, allow_compressed=True),
                    self.dir),
                'Manifest.gz')

    def test_find_top_level_manifest_from_empty_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'suba'),
                        allow_compressed=True),
                    self.dir),
                'Manifest.gz')

    def test_find_top_level_manifest_from_manifest_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'subb'),
                        allow_compressed=True),
                    self.dir),
                'Manifest.gz')

    def test_find_top_level_manifest_from_deep_manifest_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'subc', 'sub'),
                        allow_compressed=True),
                    self.dir),
                'Manifest.gz')


class TestCompressedManifestWithIgnore(TempDirTestCase):
    DIRS = ['suba', 'subb', 'subc', 'subc/sub']
    FILES = {
        'subb/Manifest': u'',
    }

    def setUp(self):
        super(TestCompressedManifestWithIgnore, self).setUp()
        with gzip.GzipFile(os.path.join(self.dir, 'Manifest.gz'), 'wb') as f:
            f.write(b'IGNORE suba\n')
            f.write(b'IGNORE subc\n')
        with gzip.GzipFile(os.path.join(self.dir, 'subc/sub/Manifest.gz'), 'wb'):
            pass

    def test_find_top_level_manifest(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        self.dir, allow_compressed=True),
                    self.dir),
                'Manifest.gz')

    def test_find_top_level_manifest_from_ignored_empty_subdir(self):
        self.assertIsNone(
                gemato.find_top_level.find_top_level_manifest(
                    os.path.join(self.dir, 'suba'),
                    allow_compressed=True))

    def test_find_top_level_manifest_from_manifest_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'subb'),
                        allow_compressed=True),
                    self.dir),
                'Manifest.gz')

    def test_find_top_level_manifest_from_deep_ignored_subdir(self):
        self.assertEqual(
                os.path.relpath(
                    gemato.find_top_level.find_top_level_manifest(
                        os.path.join(self.dir, 'subc', 'sub'),
                        allow_compressed=True),
                    self.dir),
                'subc/sub/Manifest.gz')



# gemato: exceptions
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license

class UnsupportedCompression(Exception):
    __slots__ = ['suffix']

    def __init__(self, suffix):
        super(UnsupportedCompression, self).__init__(suffix)
        self.suffix = suffix

    def __str__(self):
        return 'Unsupported compression suffix: {}'.format(self.suffix)


class UnsupportedHash(Exception):
    __slots__ = ['hash_name']

    def __init__(self, hash_name):
        super(UnsupportedHash, self).__init__(hash_name)
        self.hash_name = hash_name

    def __str__(self):
        return 'Unsupported hash name: {}'.format(self.hash_name)


class ManifestSyntaxError(Exception):
    def __init__(self, message):
        super(ManifestSyntaxError, self).__init__(message)


class ManifestIncompatibleEntry(Exception):
    __slots__ = ['e1', 'e2', 'diff']

    def __init__(self, e1, e2, diff):
        super(ManifestIncompatibleEntry, self).__init__(e1, e2, diff)
        self.e1 = e1
        self.e2 = e2
        self.diff = diff

    def __str__(self):
        msg = "Incompatible Manifest entries for {}".format(self.e1.path)
        for k, d1, d2 in self.diff:
            msg += "\n  {}: e1: {}, e2: {}".format(k, d1, d2)
        return msg


class ManifestMismatch(Exception):
    """
    An exception raised for verification failure.
    """

    __slots__ = ['path', 'entry', 'diff']

    def __init__(self, path, entry, diff):
        super(ManifestMismatch, self).__init__(path, entry, diff)
        self.path = path
        self.entry = entry
        self.diff = diff

    def __str__(self):
        msg = "Manifest mismatch for {}".format(self.path)
        for k, exp, got in self.diff:
            msg += "\n  {}: expected: {}, have: {}".format(k, exp, got)
        return msg


class ManifestCrossDevice(Exception):
    """
    An exception caused by attempting to cross filesystem boundaries.
    """

    __slots__ = ['path']

    def __init__(self, path):
        super(ManifestCrossDevice, self).__init__(path)
        self.path = path

    def __str__(self):
        return ("Path {} crosses filesystem boundaries, it must be IGNORE-d explicitly"
            .format(self.path))


class ManifestUnsignedData(Exception):
    """
    An exception caused by a Manifest file containing non-whitespace
    outside the OpenPGP-signed part.
    """

    def __str__(self):
        return "Unsigned data found in an OpenPGP signed Manifest"


class OpenPGPVerificationFailure(Exception):
    """
    An exception raised when OpenPGP verification fails.
    """

    __slots__ = ['output']

    def __init__(self, output):
        super(OpenPGPVerificationFailure, self).__init__(output)
        self.output = output

    def __str__(self):
        return "OpenPGP verification failed:\n{}".format(self.output)


class OpenPGPExpiredKeyFailure(OpenPGPVerificationFailure):
    """
    OpenPGP verification rejected because of expired key.
    """

    def __str__(self):
        return "OpenPGP signature rejected because of expired key:\n{}".format(self.output)


class OpenPGPRevokedKeyFailure(OpenPGPVerificationFailure):
    """
    OpenPGP verification rejected because of revoked key.
    """

    def __str__(self):
        return "OpenPGP signature rejected because of revoked key:\n{}".format(self.output)


class OpenPGPUnknownSigFailure(OpenPGPVerificationFailure):
    """
    OpenPGP verification rejected for unknown reason (i.e. unrecognized
    GPG status).
    """

    def __str__(self):
        return "OpenPGP signature rejected for unknown reason:\n{}".format(self.output)


class OpenPGPSigningFailure(Exception):
    """
    An exception raised when OpenPGP signing fails.
    """

    __slots__ = ['output']

    def __init__(self, output):
        super(OpenPGPSigningFailure, self).__init__(output)
        self.output = output

    def __str__(self):
        return "OpenPGP signing failed:\n{}".format(self.output)


class OpenPGPNoImplementation(Exception):
    """
    An exception raised when no supported OpenPGP implementation
    is available.
    """

    def __str__(self):
        return "No supported OpenPGP implementation found (install gnupg)"


class ManifestInvalidPath(Exception):
    """
    An exception raised when an invalid path tries to be added to
    Manifest.
    """

    __slots__ = ['path', 'detail']

    def __init__(self, path, detail):
        super(ManifestInvalidPath, self).__init__(path, detail)
        self.path = path
        self.detail = detail

    def __str__(self):
        return ("Attempting to add invalid path {} to Manifest: {} must not be {}"
                .format(self.path, self.detail[0], self.detail[1]))


class ManifestInvalidFilename(Exception):
    """
    An exception raised when an entry for invalid filename is created.
    """

    __slots__ = ['filename', 'pos']

    def __init__(self, filename, pos):
        super(ManifestInvalidFilename, self).__init__(filename, pos)
        self.filename = filename
        self.pos = pos

    def __str__(self):
        return ("Attempting to add invalid filename {!r} to Manifest: disallowed character U+{:04X} at position {}"
                .format(self.filename, ord(self.filename[self.pos]), self.pos))



# gemato: Utility functions
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license


def path_starts_with(path, prefix):
    """
    Returns True if the specified @path starts with the @prefix,
    performing component-wide comparison. Otherwise returns False.
    """
    return prefix == "" or (path + "/").startswith(prefix.rstrip("/") + "/")


def path_inside_dir(path, directory):
    """
    Returns True if the specified @path is inside @directory,
    performing component-wide comparison. Otherwise returns False.
    """
    return ((directory == "" and path != "")
            or path.rstrip("/").startswith(directory.rstrip("/") + "/"))


def throw_exception(e):
    """
    Raise the given exception. Needed for onerror= argument
    to os.walk(). Useful for other callbacks.
    """
    raise e



# gemato: Top-level Manifest finding routine
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license

import errno
import os
import os.path

import gemato.compression
import gemato.manifest


def find_top_level_manifest(path='.', allow_compressed=False):
    """
    Find top-level Manifest file that covers @path (defaults
    to the current directory). Returns the path to the Manifest
    or None.

    If @allow_compressed is true, the function allows the top-level
    Manifest to be compressed and opens all compressed files *without*
    verifying them first. It is false by default to prevent zip bombs
    and other decompression attacks.
    """

    cur_path = path
    last_found = None
    original_dev = None
    m = gemato.manifest.ManifestFile()

    root_st = os.stat('/')

    manifest_filenames = ('Manifest',)
    if allow_compressed:
        manifest_filenames = list(gemato.compression
                .get_potential_compressed_names('Manifest'))

    while True:
        st = os.stat(cur_path)

        # verify that we are not crossing device boundaries
        if original_dev is None:
            original_dev = st.st_dev
        elif original_dev != st.st_dev:
            break

        m_path = os.path.join(cur_path, 'Manifest')
        for m_name in manifest_filenames:
            m_path = os.path.join(cur_path, m_name)
            try:
                # note: this is safe for allow_compressed=False
                # since it detects compression by filename suffix
                with (gemato.compression
                        .open_potentially_compressed_path(m_path, 'r',
                            encoding='utf8')) as f:
                    fst = os.fstat(f.fileno())
                    if fst.st_dev != original_dev:
                        return last_found

                    m.load(f, verify_openpgp=False)
            except IOError as e:
                if e.errno != errno.ENOENT:
                    raise
            else:
                # check if the initial path is ignored
                relpath = os.path.relpath(path, cur_path)
                if relpath == '.':
                    relpath = ''
                fe = m.find_path_entry(relpath)
                if fe is not None and fe.tag == 'IGNORE':
                    return last_found

                last_found = m_path
                break

        # check if we reached root directory
        if st.st_dev == root_st.st_dev and st.st_ino == root_st.st_ino:
            break

        # try the parent directory
        cur_path = os.path.join(cur_path, '..')

    return last_found



# gemato: Profile behavior tests
# vim:fileencoding=utf-8
# (c) 2017 Michał Górny
# Licensed under the terms of 2-clause BSD license

import os.path

import gemato.cli
import gemato.profile
import gemato.recursiveloader

from tests.testutil import TempDirTestCase


class EbuildRepositoryTests(TempDirTestCase):
    """
    Tests for ebuild repository profiles.
    """

    PROFILE = gemato.profile.EbuildRepositoryProfile
    PROFILE_NAME = 'ebuild'
    DIRS = [
        'dev-foo',
        'dev-foo/bar',
        'dev-foo/bar/files',
        'eclass',
        'eclass/tests',
        'licenses',
        'metadata',
        'metadata/dtd',
        'metadata/glsa',
        'metadata/install-qa-check.d',
        'metadata/md5-cache',
        'metadata/md5-cache/dev-foo',
        'metadata/news',
        'metadata/news/2020-01-01-foo',
        'metadata/xml-schema',
        'profiles',
        'profiles/arch',
        'profiles/arch/foo',
        'profiles/desc',
        'profiles/updates',
    ]
    EXPECTED_MANIFESTS = [
        'dev-foo/Manifest',
        'dev-foo/bar/Manifest',
        'eclass/Manifest',
        'licenses/Manifest',
        'metadata/Manifest',
        'metadata/dtd/Manifest',
        'metadata/glsa/Manifest',
        'metadata/md5-cache/Manifest',
        'metadata/md5-cache/dev-foo/Manifest',
        'metadata/news/Manifest',
        'metadata/xml-schema/Manifest',
        'profiles/Manifest',
    ]
    EXPECTED_TYPES = {
        'header.txt': 'DATA',
        'skel.ebuild': 'DATA',
        'skel.metadata.xml': 'DATA',
        'dev-foo/metadata.xml': 'DATA',
        'dev-foo/bar/bar-1.ebuild': 'DATA',
        'dev-foo/bar/metadata.xml': 'DATA',
        'dev-foo/bar/files/test.patch': 'DATA',
        'eclass/foo.eclass': 'DATA',
        'eclass/tests/foo.sh': 'DATA',
        'licenses/foo': 'DATA',
        'metadata/layout.conf': 'DATA',
        'metadata/projects.xml': 'DATA',
        'metadata/pkg_desc_index': 'DATA',
        'metadata/dtd/foo.dtd': 'DATA',
        'metadata/glsa/glsa-202001-01.xml': 'DATA',
        'metadata/install-qa-check.d/50foo': 'DATA',
        'metadata/md5-cache/dev-foo/bar-1': 'DATA',
        'metadata/news/2020-01-01-foo/2020-01-01-foo.en.txt': 'DATA',
        'metadata/news/2020-01-01-foo/2020-01-01-foo.en.txt.asc': 'DATA',
        'metadata/xml-schema/foo.xsd': 'DATA',
        'profiles/arch.desc': 'DATA',
        'profiles/categories': 'DATA',
        'profiles/eapi': 'DATA',
        'profiles/info_pkgs': 'DATA',
        'profiles/info_vars': 'DATA',
        'profiles/license_groups': 'DATA',
        'profiles/package.mask': 'DATA',
        'profiles/profiles.desc': 'DATA',
        'profiles/repo_name': 'DATA',
        'profiles/thirdpartymirrors': 'DATA',
        'profiles/use.desc': 'DATA',
        'profiles/use.local.desc': 'DATA',
        'profiles/arch/foo/eapi': 'DATA',
        'profiles/arch/foo/parent': 'DATA',
        'profiles/desc/foo.desc': 'DATA',
        'profiles/updates/1Q-2020': 'DATA',
    }
    FILES = dict.fromkeys(list(EXPECTED_TYPES) + [
        'metadata/timestamp',
        'metadata/timestamp.chk',
        'metadata/timestamp.commit',
        'metadata/timestamp.x',
    ], u'')

    EXPECTED_IGNORE = [
        'distfiles',
        'local',
        'lost+found',
        'packages',
        'metadata/timestamp',
        'metadata/timestamp.chk',
        'metadata/timestamp.commit',
        'metadata/timestamp.x',
        'metadata/dtd/timestamp.chk',
        'metadata/dtd/timestamp.commit',
        'metadata/glsa/timestamp.chk',
        'metadata/glsa/timestamp.commit',
        'metadata/news/timestamp.chk',
        'metadata/news/timestamp.commit',
        'metadata/xml-schema/timestamp.chk',
        'metadata/xml-schema/timestamp.commit',
    ]

    def test_get_entry_type_for_path(self):
        p = self.PROFILE()
        for f, expt in self.EXPECTED_TYPES.items():
            self.assertEqual(
                    p.get_entry_type_for_path(f),
                    expt,
                    "type mismatch for {}".format(f))

    def test_update_entries_for_directory(self):
        m = gemato.recursiveloader.ManifestRecursiveLoader(
                os.path.join(self.dir, 'Manifest'),
                hashes=['SHA256', 'SHA512'],
                allow_create=True,
                profile=self.PROFILE())
        m.update_entries_for_directory('')
        for f, expt in self.EXPECTED_TYPES.items():
            self.assertEqual(
                    m.find_path_entry(f).tag,
                    expt,
                    "type mismatch for {}".format(f))
        for f in self.EXPECTED_MANIFESTS:
            self.assertEqual(m.find_path_entry(f).tag, 'MANIFEST',
                    "type mismatch for {}".format(f))
        for f in self.EXPECTED_IGNORE:
            self.assertIsNotNone(m.find_path_entry(f),
                    "missing IGNORE entry for {}".format(f))
            self.assertEqual(m.find_path_entry(f).tag, 'IGNORE',
                    "type mismatch for {}".format(f))
        m.save_manifests()
        m.assert_directory_verifies('')
        return m

    def test_regression_top_level_ignore_in_all_manifests(self):
        assert 'distfiles' in self.EXPECTED_IGNORE[0]
        assert 'dev-foo/Manifest' in self.EXPECTED_MANIFESTS

        m = gemato.recursiveloader.ManifestRecursiveLoader(
                os.path.join(self.dir, 'Manifest'),
                hashes=['SHA256', 'SHA512'],
                allow_create=True,
                profile=self.PROFILE())
        m.update_entries_for_directory('')

        self.assertIsNone(m.find_path_entry('dev-foo/distfiles'))

    def test_set_loader_options(self):
        m = gemato.recursiveloader.ManifestRecursiveLoader(
                os.path.join(self.dir, 'Manifest'),
                profile=self.PROFILE(),
                allow_create=True)
        self.assertIsNotNone(m.hashes)
        self.assertTrue(m.sort)
        self.assertIsNotNone(m.compress_watermark)
        self.assertIsNotNone(m.compress_format)

    def test_cli_update(self):
        self.assertEqual(
                gemato.cli.main(['gemato', 'create',
                    '--profile', self.PROFILE_NAME,
                    self.dir]),
                0)

        m = gemato.recursiveloader.ManifestRecursiveLoader(
                os.path.join(self.dir, 'Manifest'))
        for f, expt in self.EXPECTED_TYPES.items():
            self.assertEqual(
                    m.find_path_entry(f).tag,
                    expt,
                    "type mismatch for {}".format(f))
        for f in self.EXPECTED_MANIFESTS:
            self.assertEqual(m.find_path_entry(f).tag, 'MANIFEST',
                    "type mismatch for {}".format(f))

        self.assertEqual(
                gemato.cli.main(['gemato', 'verify', self.dir]),
                0)

        return m


class BackwardsCompatEbuildRepositoryTests(EbuildRepositoryTests):
    PROFILE = gemato.profile.BackwardsCompatEbuildRepositoryProfile
    PROFILE_NAME = 'old-ebuild'

    def __init__(self, *args, **kwargs):
        self.EXPECTED_TYPES = self.EXPECTED_TYPES.copy()
        self.EXPECTED_TYPES.update({
            'dev-foo/bar/bar-1.ebuild': 'EBUILD',
            'dev-foo/bar/metadata.xml': 'MISC',
            'dev-foo/bar/files/test.patch': 'AUX',
        })
        super(BackwardsCompatEbuildRepositoryTests, self).__init__(
                *args, **kwargs)

    def test_update_entries_for_directory(self):
        m = (super(BackwardsCompatEbuildRepositoryTests, self)
                .test_update_entries_for_directory())
        self.assertEqual(
                m.find_path_entry('dev-foo/bar/files/test.patch').path,
                'files/test.patch')
        self.assertEqual(
                m.find_path_entry('dev-foo/bar/files/test.patch').aux_path,
                'test.patch')

    def test_cli_update(self):
        m = (super(BackwardsCompatEbuildRepositoryTests, self)
                .test_cli_update())
        self.assertEqual(
                m.find_path_entry('dev-foo/bar/files/test.patch').path,
                'files/test.patch')
        self.assertEqual(
                m.find_path_entry('dev-foo/bar/files/test.patch').aux_path,
                'test.patch')

    def test_compression(self):
        """
        Test that package directory Manifests are not compressed.
        """

        m = gemato.recursiveloader.ManifestRecursiveLoader(
                os.path.join(self.dir, 'Manifest'),
                hashes=['SHA256', 'SHA512'],
                compress_watermark=0,
                allow_create=True,
                profile=self.PROFILE())
        m.update_entries_for_directory('')
        m.save_manifests()

        for mpath in self.EXPECTED_MANIFESTS:
            # package manifest should be left uncompressed
            if mpath == 'dev-foo/bar/Manifest':
                self.assertTrue(os.path.exists(os.path.join(
                    self.dir, mpath)))
            else:
                self.assertTrue(os.path.exists(os.path.join(
                    self.dir, mpath + '.gz')))
                self.assertFalse(os.path.exists(os.path.join(
                    self.dir, mpath)))

    def test_cli_compression(self):
        self.assertEqual(
                gemato.cli.main(['gemato', 'create',
                    '--profile', self.PROFILE_NAME,
                    '--compress-watermark=0',
                    self.dir]),
                0)

        for mpath in self.EXPECTED_MANIFESTS:
            # package manifest should be left uncompressed
            if mpath == 'dev-foo/bar/Manifest':
                self.assertTrue(os.path.exists(os.path.join(
                    self.dir, mpath)))
            else:
                self.assertTrue(os.path.exists(os.path.join(
                    self.dir, mpath + '.gz')))
                self.assertFalse(os.path.exists(os.path.join(
                    self.dir, mpath)))
